/* pango-script-lang-table.h:
 * 
 * Generated by ../tools/gen-script-for-lang.c
 * Date: 2019-06-06
 * Source: fontconfig-2.13.1
 * 
 * Do not edit.
 */
typedef struct _PangoScriptForLang {
  const char lang[9];
  GUnicodeScript scripts[3];
} PangoScriptForLang;

static const PangoScriptForLang pango_script_for_lang[] = {
  { "aa",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "ab",       { G_UNICODE_SCRIPT_CYRILLIC/*90*/ } },
  { "af",       { G_UNICODE_SCRIPT_LATIN/*69*/ } },
  { "ak",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "am",       { G_UNICODE_SCRIPT_ETHIOPIC/*264*/ } },
  { "an",       { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "ar",       { G_UNICODE_SCRIPT_ARABIC/*36*/ } },
  { "as",       { G_UNICODE_SCRIPT_BENGALI/*64*/ } },
  { "ast",      { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "av",       { G_UNICODE_SCRIPT_CYRILLIC/*67*/ } },
  { "ay",       { G_UNICODE_SCRIPT_LATIN/*60*/ } },
  { "az-az",    { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "az-ir",    { G_UNICODE_SCRIPT_ARABIC/*38*/ } },
  { "ba",       { G_UNICODE_SCRIPT_CYRILLIC/*82*/ } },
  { "be",       { G_UNICODE_SCRIPT_CYRILLIC/*68*/ } },
  { "ber-dz",   { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "ber-ma",   { G_UNICODE_SCRIPT_TIFINAGH/*32*/ } },
  { "bg",       { G_UNICODE_SCRIPT_CYRILLIC/*60*/ } },
  { "bh",       { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "bho",      { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "bi",       { G_UNICODE_SCRIPT_LATIN/*58*/ } },
  { "bin",      { G_UNICODE_SCRIPT_LATIN/*76*/ } },
  { "bm",       { G_UNICODE_SCRIPT_LATIN/*60*/ } },
  { "bn",       { G_UNICODE_SCRIPT_BENGALI/*63*/ } },
  { "bo",       { G_UNICODE_SCRIPT_TIBETAN/*95*/ } },
  { "br",       { G_UNICODE_SCRIPT_LATIN/*64*/ } },
  { "brx",      { G_UNICODE_SCRIPT_DEVANAGARI/*78*/ } },
  { "bs",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "bua",      { G_UNICODE_SCRIPT_CYRILLIC/*70*/ } },
  { "byn",      { G_UNICODE_SCRIPT_ETHIOPIC/*255*/ } },
  { "ca",       { G_UNICODE_SCRIPT_LATIN/*74*/ } },
  { "ce",       { G_UNICODE_SCRIPT_CYRILLIC/*67*/ } },
  { "ch",       { G_UNICODE_SCRIPT_LATIN/*58*/ } },
  { "chm",      { G_UNICODE_SCRIPT_CYRILLIC/*76*/ } },
  { "chr",      { G_UNICODE_SCRIPT_CHEROKEE/*85*/ } },
  { "co",       { G_UNICODE_SCRIPT_LATIN/*84*/ } },
  { "crh",      { G_UNICODE_SCRIPT_LATIN/*68*/ } },
  { "cs",       { G_UNICODE_SCRIPT_LATIN/*82*/ } },
  { "csb",      { G_UNICODE_SCRIPT_LATIN/*74*/ } },
  { "cu",       { G_UNICODE_SCRIPT_CYRILLIC/*103*/ } },
  { "cv",       { G_UNICODE_SCRIPT_CYRILLIC/*72*/, G_UNICODE_SCRIPT_LATIN/*2*/ } },
  { "cy",       { G_UNICODE_SCRIPT_LATIN/*78*/ } },
  { "da",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "de",       { G_UNICODE_SCRIPT_LATIN/*59*/ } },
  { "doi",      { G_UNICODE_SCRIPT_DEVANAGARI/*81*/ } },
  { "dv",       { G_UNICODE_SCRIPT_THAANA/*49*/ } },
  { "dz",       { G_UNICODE_SCRIPT_TIBETAN/*95*/ } },
  { "ee",       { G_UNICODE_SCRIPT_LATIN/*96*/ } },
  { "el",       { G_UNICODE_SCRIPT_GREEK/*69*/ } },
  { "en",       { G_UNICODE_SCRIPT_LATIN/*72*/ } },
  { "eo",       { G_UNICODE_SCRIPT_LATIN/*64*/ } },
  { "es",       { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "et",       { G_UNICODE_SCRIPT_LATIN/*64*/ } },
  { "eu",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "fa",       { G_UNICODE_SCRIPT_ARABIC/*38*/ } },
  { "fat",      { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "ff",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "fi",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "fil",      { G_UNICODE_SCRIPT_LATIN/*84*/ } },
  { "fj",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "fo",       { G_UNICODE_SCRIPT_LATIN/*68*/ } },
  { "fr",       { G_UNICODE_SCRIPT_LATIN/*84*/ } },
  { "fur",      { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "fy",       { G_UNICODE_SCRIPT_LATIN/*75*/ } },
  { "ga",       { G_UNICODE_SCRIPT_LATIN/*80*/ } },
  { "gd",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "gez",      { G_UNICODE_SCRIPT_ETHIOPIC/*218*/ } },
  { "gl",       { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "gn",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "gu",       { G_UNICODE_SCRIPT_GUJARATI/*68*/ } },
  { "gv",       { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "ha",       { G_UNICODE_SCRIPT_LATIN/*60*/ } },
  { "haw",      { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "he",       { G_UNICODE_SCRIPT_HEBREW/*27*/ } },
  { "hi",       { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "hne",      { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "ho",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "hr",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "hsb",      { G_UNICODE_SCRIPT_LATIN/*72*/ } },
  { "ht",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "hu",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "hy",       { G_UNICODE_SCRIPT_ARMENIAN/*77*/ } },
  { "hz",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "ia",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "id",       { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "ie",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "ig",       { G_UNICODE_SCRIPT_LATIN/*58*/ } },
  { "ii",       { G_UNICODE_SCRIPT_YI/*1165*/ } },
  { "ik",       { G_UNICODE_SCRIPT_CYRILLIC/*68*/ } },
  { "io",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "is",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "it",       { G_UNICODE_SCRIPT_LATIN/*72*/ } },
  { "iu",       { G_UNICODE_SCRIPT_CANADIAN_ABORIGINAL/*161*/ } },
  { "ja",       { G_UNICODE_SCRIPT_HAN/*2134*/, G_UNICODE_SCRIPT_KATAKANA/*88*/, G_UNICODE_SCRIPT_HIRAGANA/*85*/ } },
  { "jv",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "ka",       { G_UNICODE_SCRIPT_GEORGIAN/*33*/ } },
  { "kaa",      { G_UNICODE_SCRIPT_CYRILLIC/*78*/ } },
  { "kab",      { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "ki",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "kj",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "kk",       { G_UNICODE_SCRIPT_CYRILLIC/*77*/ } },
  { "kl",       { G_UNICODE_SCRIPT_LATIN/*81*/ } },
  { "km",       { G_UNICODE_SCRIPT_KHMER/*63*/ } },
  { "kn",       { G_UNICODE_SCRIPT_KANNADA/*70*/ } },
  { "ko",       { G_UNICODE_SCRIPT_HANGUL/*2442*/ } },
  { "kok",      { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "kr",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "ks",       { G_UNICODE_SCRIPT_ARABIC/*33*/ } },
  { "ku-am",    { G_UNICODE_SCRIPT_CYRILLIC/*64*/ } },
  { "ku-iq",    { G_UNICODE_SCRIPT_ARABIC/*32*/ } },
  { "ku-ir",    { G_UNICODE_SCRIPT_ARABIC/*32*/ } },
  { "ku-tr",    { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "kum",      { G_UNICODE_SCRIPT_CYRILLIC/*66*/ } },
  { "kv",       { G_UNICODE_SCRIPT_CYRILLIC/*70*/ } },
  { "kw",       { G_UNICODE_SCRIPT_LATIN/*64*/ } },
  { "kwm",      { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "ky",       { G_UNICODE_SCRIPT_CYRILLIC/*70*/ } },
  { "la",       { G_UNICODE_SCRIPT_LATIN/*68*/ } },
  { "lah",      { G_UNICODE_SCRIPT_ARABIC/*27*/ } },
  { "lb",       { G_UNICODE_SCRIPT_LATIN/*75*/ } },
  { "lez",      { G_UNICODE_SCRIPT_CYRILLIC/*67*/ } },
  { "lg",       { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "li",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "ln",       { G_UNICODE_SCRIPT_LATIN/*78*/ } },
  { "lo",       { G_UNICODE_SCRIPT_LAO/*55*/ } },
  { "lt",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "lv",       { G_UNICODE_SCRIPT_LATIN/*78*/ } },
  { "mai",      { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "mg",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "mh",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "mi",       { G_UNICODE_SCRIPT_LATIN/*64*/ } },
  { "mk",       { G_UNICODE_SCRIPT_CYRILLIC/*42*/ } },
  { "ml",       { G_UNICODE_SCRIPT_MALAYALAM/*68*/ } },
  { "mn-cn",    { G_UNICODE_SCRIPT_MONGOLIAN/*130*/ } },
  { "mn-mn",    { G_UNICODE_SCRIPT_CYRILLIC/*70*/ } },
  { "mni",      { G_UNICODE_SCRIPT_BENGALI/*75*/ } },
  { "mo",       { G_UNICODE_SCRIPT_CYRILLIC/*66*/, G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "mr",       { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "ms",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "mt",       { G_UNICODE_SCRIPT_LATIN/*72*/ } },
  { "my",       { G_UNICODE_SCRIPT_MYANMAR/*48*/ } },
  { "na",       { G_UNICODE_SCRIPT_LATIN/*60*/ } },
  { "nb",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "nds",      { G_UNICODE_SCRIPT_LATIN/*59*/ } },
  { "ne",       { G_UNICODE_SCRIPT_DEVANAGARI/*70*/ } },
  { "ng",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "nl",       { G_UNICODE_SCRIPT_LATIN/*82*/ } },
  { "nn",       { G_UNICODE_SCRIPT_LATIN/*76*/ } },
  { "no",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "nqo",      { G_UNICODE_SCRIPT_NKO/*59*/ } },
  { "nr",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "nso",      { G_UNICODE_SCRIPT_LATIN/*58*/ } },
  { "nv",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "ny",       { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "oc",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "om",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "or",       { G_UNICODE_SCRIPT_ORIYA/*68*/ } },
  { "os",       { G_UNICODE_SCRIPT_CYRILLIC/*66*/ } },
  { "ota",      { G_UNICODE_SCRIPT_ARABIC/*37*/ } },
  { "pa",       { G_UNICODE_SCRIPT_GURMUKHI/*63*/ } },
  { "pa-pk",    { G_UNICODE_SCRIPT_ARABIC/*27*/ } },
  { "pap-an",   { G_UNICODE_SCRIPT_LATIN/*72*/ } },
  { "pap-aw",   { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "pl",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "ps-af",    { G_UNICODE_SCRIPT_ARABIC/*49*/ } },
  { "ps-pk",    { G_UNICODE_SCRIPT_ARABIC/*49*/ } },
  { "pt",       { G_UNICODE_SCRIPT_LATIN/*82*/ } },
  { "qu",       { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "quz",      { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "rm",       { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "rn",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "ro",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "ru",       { G_UNICODE_SCRIPT_CYRILLIC/*66*/ } },
  { "rw",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "sa",       { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "sah",      { G_UNICODE_SCRIPT_CYRILLIC/*76*/ } },
  { "sat",      { G_UNICODE_SCRIPT_DEVANAGARI/*68*/ } },
  { "sc",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "sco",      { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "sd",       { G_UNICODE_SCRIPT_ARABIC/*54*/ } },
  { "se",       { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "sel",      { G_UNICODE_SCRIPT_CYRILLIC/*66*/ } },
  { "sg",       { G_UNICODE_SCRIPT_LATIN/*72*/ } },
  { "sh",       { G_UNICODE_SCRIPT_CYRILLIC/*94*/, G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "shs",      { G_UNICODE_SCRIPT_LATIN/*46*/ } },
  { "si",       { G_UNICODE_SCRIPT_SINHALA/*73*/ } },
  { "sid",      { G_UNICODE_SCRIPT_ETHIOPIC/*281*/ } },
  { "sk",       { G_UNICODE_SCRIPT_LATIN/*86*/ } },
  { "sl",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "sm",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "sma",      { G_UNICODE_SCRIPT_LATIN/*60*/ } },
  { "smj",      { G_UNICODE_SCRIPT_LATIN/*60*/ } },
  { "smn",      { G_UNICODE_SCRIPT_LATIN/*68*/ } },
  { "sms",      { G_UNICODE_SCRIPT_LATIN/*80*/ } },
  { "sn",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "so",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "sq",       { G_UNICODE_SCRIPT_LATIN/*56*/ } },
  { "sr",       { G_UNICODE_SCRIPT_CYRILLIC/*60*/ } },
  { "ss",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "st",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "su",       { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "sv",       { G_UNICODE_SCRIPT_LATIN/*68*/ } },
  { "sw",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "syr",      { G_UNICODE_SCRIPT_SYRIAC/*45*/ } },
  { "ta",       { G_UNICODE_SCRIPT_TAMIL/*48*/ } },
  { "te",       { G_UNICODE_SCRIPT_TELUGU/*70*/ } },
  { "tg",       { G_UNICODE_SCRIPT_CYRILLIC/*78*/ } },
  { "th",       { G_UNICODE_SCRIPT_THAI/*73*/ } },
  { "ti-er",    { G_UNICODE_SCRIPT_ETHIOPIC/*255*/ } },
  { "ti-et",    { G_UNICODE_SCRIPT_ETHIOPIC/*281*/ } },
  { "tig",      { G_UNICODE_SCRIPT_ETHIOPIC/*221*/ } },
  { "tk",       { G_UNICODE_SCRIPT_LATIN/*68*/ } },
  { "tl",       { G_UNICODE_SCRIPT_LATIN/*84*/ } },
  { "tn",       { G_UNICODE_SCRIPT_LATIN/*58*/ } },
  { "to",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "tr",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "ts",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "tt",       { G_UNICODE_SCRIPT_CYRILLIC/*76*/ } },
  { "tw",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "ty",       { G_UNICODE_SCRIPT_LATIN/*64*/ } },
  { "tyv",      { G_UNICODE_SCRIPT_CYRILLIC/*70*/ } },
  { "ug",       { G_UNICODE_SCRIPT_ARABIC/*33*/ } },
  { "uk",       { G_UNICODE_SCRIPT_CYRILLIC/*72*/ } },
  { "und-zmth", { G_UNICODE_SCRIPT_LATIN/*53*/, G_UNICODE_SCRIPT_GREEK/*51*/ } },
  { "und-zsye", { 0 } },
  { "ur",       { G_UNICODE_SCRIPT_ARABIC/*27*/ } },
  { "uz",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "ve",       { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "vi",       { G_UNICODE_SCRIPT_LATIN/*186*/ } },
  { "vo",       { G_UNICODE_SCRIPT_LATIN/*54*/ } },
  { "vot",      { G_UNICODE_SCRIPT_LATIN/*62*/ } },
  { "wa",       { G_UNICODE_SCRIPT_LATIN/*70*/ } },
  { "wal",      { G_UNICODE_SCRIPT_ETHIOPIC/*281*/ } },
  { "wen",      { G_UNICODE_SCRIPT_LATIN/*76*/ } },
  { "wo",       { G_UNICODE_SCRIPT_LATIN/*66*/ } },
  { "xh",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "yap",      { G_UNICODE_SCRIPT_LATIN/*58*/ } },
  { "yi",       { G_UNICODE_SCRIPT_HEBREW/*27*/ } },
  { "yo",       { G_UNICODE_SCRIPT_LATIN/*114*/ } },
  { "za",       { G_UNICODE_SCRIPT_LATIN/*52*/ } },
  { "zh-cn",    { G_UNICODE_SCRIPT_HAN/*6763*/ } },
  { "zh-hk",    { G_UNICODE_SCRIPT_HAN/*1083*/ } },
  { "zh-mo",    { G_UNICODE_SCRIPT_HAN/*1083*/ } },
  { "zh-sg",    { G_UNICODE_SCRIPT_HAN/*6763*/ } },
  { "zh-tw",    { G_UNICODE_SCRIPT_HAN/*13063*/ } },
  { "zu",       { G_UNICODE_SCRIPT_LATIN/*52*/ } }
};
